void calcSteeringPID(void)
{
    //Proportional only
    pValue = Set.Kp * steerAngleError;

    Serial.println(pwmDrive);
    Serial.println(steerAngleError);
    Serial.println(pValue);
    
    pwmDrive = (int)pValue;

    errorAbs = abs(steerAngleError);
    float newMax = 0;

    if (errorAbs < Set.MotorSlowDriveDegrees)
    {
        newMax = (errorAbs * highLowPerDeg) + Set.lowPWM;
    }
    else newMax = Set.highPWM;

    //add min throttle factor so no delay from motor resistance.
    if (pwmDrive < 0) pwmDrive -= Set.minPWM;
    else if (pwmDrive > 0) pwmDrive += Set.minPWM;

    //Serial.print(newMax); //The actual steering angle in degrees
    //Serial.print(",");

 //limit the pwm drive
    if (pwmDrive > newMax) pwmDrive = newMax;
    if (pwmDrive < -newMax) pwmDrive = -newMax;

    if (Set.MotorDriveDirection) pwmDrive *= -1;
    if (Set.debugmode) { Serial.print("PWM for output: ");  Serial.println(pwmDrive); }
}

//-------------------------------------------------------------------------------------------------
// select output Driver
//---------------------------------------------------------------------
void motorDrive(void) 
{
  switch (Set.output_type) {
    case 1:
      motorDrive_Cytron();
      break;
    case 2:
      motorDrive_IBT_Mot();
      break;
    case 3:
      motorDrive_IBT_PWM();
      break;
    case 4:
      motorDrive_IBT_Danfoss();
      break;
    case 5:
      motorDrive_Keya_v2020();
      break;
    case 6:
      motorDrive_Keya_v2021();
      break;
    case 7:
      claas();
      break;
    case 8:
      motorDrive_Keya_Serial_v2021();
      break;
    default:
      // if nothing else matches no Output
    break;
  }
}
 
//---------------------------------------------------------------------
// Used with Cytron MD30C Driver
// Steering Motor
// Dir + PWM Signal
//---------------------------------------------------------------------
void motorDrive_Cytron(void) 
  {    
    pwmDisplay = pwmDrive;
    if (pwmDrive >= 0) ledcWrite(1, 255);  // channel 1 = DIR_PIN  //set the correct direction
    else   
    {
      ledcWrite(1, 0);  // channel 1 = DIR_PIN 
      pwmDrive = -1 * pwmDrive;  
   }
    ledcWrite(0, pwmDrive);  // channel 0 = PWM_PIN
  }


//---------------------------------------------------------------------
// Used with IBT 2  Driver
// Steering Motor
// PWM Left + PWM Right Signal
// Same Code as the PWM Valve
//---------------------------------------------------------------------
void motorDrive_IBT_Mot(void) 
  {   
   if (steerEnable == false) pwmDrive=0;
   
    pwmDisplay = pwmDrive; 
  
  if (pwmDrive > 0)
    {
      ledcWrite(0, pwmDrive);  // channel 0 = PWM_PIN
      ledcWrite(1, 0);
      //digitalWrite(DIR_PIN, LOW);
    }
    
  if (pwmDrive <= 0)
    {
      pwmDrive = -1 * pwmDrive;  
      ledcWrite(1, pwmDrive);  // channel 1 = DIR_PIN
      ledcWrite(0, 0);
      //digitalWrite(PWM_PIN, LOW);
    } 
  }

//---------------------------------------------------------------------
// Used with 2-Coil PWM Valves 
// No coil powered = Center
// Coil 1 steer right connected to PWM_PIN 11
// Coil 2 steer left  connected to DIR_PIN 10
//---------------------------------------------------------------------
void motorDrive_IBT_PWM(void)
 { 
  if (steerEnable == false) pwmDrive=0;
   
  pwmDisplay = pwmDrive; 
  
  if (pwmDrive > 0)
    {
      ledcWrite(0, pwmDrive);  // channel 0 = PWM_PIN
      ledcWrite(1, 0);
      //digitalWrite(DIR_PIN, LOW);
    }
    
  if (pwmDrive <= 0)
    {
      pwmDrive = -1 * pwmDrive;  
      ledcWrite(1, pwmDrive);  // channel 1 = DIR_PIN
      ledcWrite(0, 0);
      //digitalWrite(PWM_PIN, LOW);
    }
 }  

//---------------------------------------------------------------------
// Danfoss: PWM 25% On = Left Position max  (below Valve=Center)
// Danfoss: PWM 50% On = Center Position
// Danfoss: PWM 75% On = Right Position max (above Valve=Center)
//---------------------------------------------------------------------
void motorDrive_IBT_Danfoss(void) 
  { 
    if (pwmDrive >  250) pwmDrive =  250; 
    if (pwmDrive < -250) pwmDrive = -250;
  
 /* if (steerEnable) digitalWrite(DIR_PIN, HIGH); // turn on /off Power
  else 
    {
      digitalWrite(DIR_PIN, LOW);  
      pwmDrive=0;
    }*/
  if (steerEnable) ledcWrite(1, 255);  // channel 1 = DIR_PIN // turn on /off Power
  else 
    {
      ledcWrite(1, 0);
      pwmDrive=0;
    }
    pwmDrive = pwmDrive / 4;  
    pwmOut = pwmDrive+ 128;  // add Center Pos.
    pwmDisplay = pwmOut;
    ledcWrite(0, pwmOut);  // channel 0 = PWM_PIN
  }


//--------------------------------------------------------------------
// Keya Motors http://www.dcmotorkeya.com/ manufactured till 2020
//---------------------------------------------------------------------
void motorDrive_Keya_v2020(void) 

  { 
    pwmDrive = pwmDrive * 4; // Gain is at about 30

    if (pwmDrive > 1000) pwmDrive = 1000;
    if (pwmDrive < -1000) pwmDrive = -1000;

    byte off[] = {0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  };
    byte hold[] = {0x11, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  };

    byte hi, lo;
    byte steer[] =   {0x11, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  };

    if (pwmDrive > 0)
    {
      steer[4] = 0x00;
      steer[5] = 0x00;
    }
    else
    {
      steer[4] = 0xFF;
      steer[5] = 0xFF;
    }

    hi =  ( (pwmDrive) >> (8) ); // keep upper 8 bits
    lo = ( (pwmDrive) & (0xff) ); // keep lower 8 bits

    steer[6] = hi;
    steer[7] = lo;

   if (steerEnable == true && pwmDrive != 0) { 
//    Serial.println(pwmDrive);
    UDPToWheel.beginPacket(Set.ipDestinationWheel, Set.portSendToWheel);
    UDPToWheel.write(steer, sizeof(steer));
    UDPToWheel.endPacket();
   }



// ********************** OFF *************************

   if (steerEnable == false) { 
    pwmDrive=0;
 //   Serial.println("off");
    UDPToWheel.beginPacket(Set.ipDestinationWheel, Set.portSendToWheel);
    UDPToWheel.write(off, sizeof(off));
    UDPToWheel.endPacket();
   }

// ********************** CENTER *************************

   if (steerEnable == true && pwmDrive == 0) { 
//   if (steerEnable == true && pwmDrive <= 1 && pwmDrive >= -1) { 
//    Serial.println("hold");
    UDPToWheel.beginPacket(Set.ipDestinationWheel, Set.portSendToWheel);
    UDPToWheel.write(hold, sizeof(hold));
    UDPToWheel.endPacket();
    }

  } 


//--------------------------------------------------------------------
// Keya Motors http://www.dcmotorkeya.com/ manufactured since 2021
//---------------------------------------------------------------------
void motorDrive_Keya_v2021(void) 

  { 
    pwmDrive = pwmDrive * 4; // Gain is at about 30

    if (pwmDrive > 1000) pwmDrive = 1000;
    if (pwmDrive < -1000) pwmDrive = -1000;

    byte off[] = {0xAC, 0xEC, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00  };
    byte hold[] = {0xAC, 0xEC, 0x11, 0x00, 0x00, 0x00, 0x00, 0x00  };

    byte hi, lo, ch;
    byte steer[] =   {0xAC, 0xEC, 0x11, 0x00, 0x00, 0x00, 0x00, 0x00  };

    if (pwmDrive > 0)
    {
      steer[3] = 0x00;
      steer[4] = 0x00;
    }
    else if (pwmDrive < 0)
    {
      steer[3] = 0xFF;
      steer[4] = 0xFF;
    } else 
    {
      steer[3] = 0x00;
      steer[4] = 0x00;
    }
    

    hi =  ( (pwmDrive) >> (8) ); // keep upper 8 bits
    lo = ( (pwmDrive) & (0xff) ); // keep lower 8 bits

    steer[5] = hi;
    steer[6] = lo;

    ch =   steer[0] + steer[1] + steer[2] + steer[3] + steer[4] + steer[5] + steer[6]; // checksum BEARBEITUNG NOTWENDIG !!!!!
    
    steer[7] = ch;
    
   if (steerEnable == true && pwmDrive != 0) { 
 //   Serial.println(pwmDrive);
 //   Serial.println(steer[7]);
    UDPToWheel.beginPacket(Set.ipDestinationWheel, Set.portSendToWheel);
    UDPToWheel.write(steer, sizeof(steer));
    UDPToWheel.endPacket();
   }



// ********************** OFF *************************

   if (steerEnable == false) { 
    pwmDrive=0;
/*    Serial.println(steer[0]);
    Serial.println(steer[1]);
    Serial.println(steer[2]);
    Serial.println(steer[3]);
    Serial.println(steer[4]);
    Serial.println(steer[5]);
    Serial.println(steer[6]);
    Serial.println(steer[7]);
 */   
    
    
    UDPToWheel.beginPacket(Set.ipDestinationWheel, Set.portSendToWheel);
    UDPToWheel.write(steer, sizeof(steer));
//    UDPToWheel.write(off, sizeof(off));
    UDPToWheel.endPacket();
   }

// ********************** CENTER *************************

   if (steerEnable == true && pwmDrive == 0) { 
//   if (steerEnable == true && pwmDrive <= 1 && pwmDrive >= -1) { 
//    Serial.println("hold");
    UDPToWheel.beginPacket(Set.ipDestinationWheel, Set.portSendToWheel);
    UDPToWheel.write(hold, sizeof(hold));
    UDPToWheel.endPacket();
    }

  }  

//---------------------------------------------------------------------
// Used with Claas 2-Coil PWM Valves 
// Coil 1 steer right
// Coil 2 steer left
//---------------------------------------------------------------------
void claas(void)
 { 

  
  if (steerEnable == false) pwmDrive=0;
   
  pwmDisplay = pwmDrive; 

//    Serial.println(pwmDrive);

//    pinMode(32, OUTPUT);
//    pinMode(33, OUTPUT);
//    digitalWrite(32, HIGH);
//    digitalWrite(33, HIGH);
//
//    steerEnable = false;

    
  if (pValue >= 20 && steerEnable == true)
    {
      pinMode(32, OUTPUT);
      pinMode(33, OUTPUT);

      digitalWrite(32, LOW);
      digitalWrite(33, HIGH);
//      Serial.println("Rechts"); 

    } else if (pValue <= -20 && steerEnable == true)

    {
      pinMode(32, OUTPUT);
      pinMode(33, OUTPUT);

      digitalWrite(33, LOW);
      digitalWrite(32, HIGH);
//      Serial.println("Links");      

    } else {
      pinMode(32, OUTPUT);
      pinMode(33, OUTPUT);

      digitalWrite(33, HIGH);
      digitalWrite(32, HIGH);
//      Serial.println("Gerade");      
      
    }
 
 }  



//--------------------------------------------------------------------
// Keya Motors http://www.dcmotorkeya.com/ manufactured since 2021 - DATA VIA SERIAL
// For ESP32 use Pin 16 and 17
//---------------------------------------------------------------------
void motorDrive_Keya_Serial_v2021(void) 

  { 
    pwmDrive = pwmDrive * 4; // Gain is at about 30

    if (pwmDrive > 1000) pwmDrive = 1000;
    if (pwmDrive < -1000) pwmDrive = -1000;

    byte off[] = {0xAC, 0xEC, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00  };
    byte hold[] = {0xAC, 0xEC, 0x11, 0x00, 0x00, 0x00, 0x00, 0x00  };

    byte hi, lo, ch;
    byte steer[] =   {0xAC, 0xEC, 0x11, 0x00, 0x00, 0x00, 0x00, 0x00  };

    if (pwmDrive > 0)
    {
      steer[3] = 0x00;
      steer[4] = 0x00;
    }
    else if (pwmDrive < 0)
    {
      steer[3] = 0xFF;
      steer[4] = 0xFF;
    } else 
    {
      steer[3] = 0x00;
      steer[4] = 0x00;
    }
    

    hi =  ( (pwmDrive) >> (8) ); // keep upper 8 bits
    lo = ( (pwmDrive) & (0xff) ); // keep lower 8 bits

    steer[5] = hi;
    steer[6] = lo;

    ch =   steer[0] + steer[1] + steer[2] + steer[3] + steer[4] + steer[5] + steer[6]; // checksum BEARBEITUNG NOTWENDIG !!!!!
    
    steer[7] = ch;
    
   if (steerEnable == true && pwmDrive != 0) { 
 //   Serial.println(pwmDrive);
 //   Serial.println(steer[7]);
    Serial2.write(steer, sizeof(steer));

    Serial.print(steer[0], HEX);
    Serial.print(" ");
    Serial.print(steer[1], HEX);
    Serial.print(" ");
    Serial.print(steer[2], HEX);
    Serial.print(" ");
    Serial.print(steer[3], HEX);
    Serial.print(" ");
    Serial.print(steer[4], HEX);
    Serial.print(" ");
    Serial.print(steer[5], HEX);
    Serial.print(" ");
    Serial.print(steer[6], HEX);
    Serial.print(" ");
    Serial.println(steer[7], HEX);



}



// ********************** OFF *************************

   if (steerEnable == false) { 
    pwmDrive=0;
    
    Serial2.write(steer, sizeof(steer));

    Serial.print(steer[0], HEX);
    Serial.print(" ");
    Serial.print(steer[1], HEX);
    Serial.print(" ");
    Serial.print(steer[2], HEX);
    Serial.print(" ");
    Serial.print(steer[3], HEX);
    Serial.print(" ");
    Serial.print(steer[4], HEX);
    Serial.print(" ");
    Serial.print(steer[5], HEX);
    Serial.print(" ");
    Serial.print(steer[6], HEX);
    Serial.print(" ");
    Serial.println(steer[7], HEX);

   }

// ********************** CENTER *************************

   if (steerEnable == true && pwmDrive == 0) { 
//   if (steerEnable == true && pwmDrive <= 1 && pwmDrive >= -1) { 
//    Serial.println("hold");

    
    Serial2.write(hold, sizeof(hold));

    Serial.print(steer[0], HEX);
    Serial.print(" ");
    Serial.print(steer[1], HEX);
    Serial.print(" ");
    Serial.print(steer[2], HEX);
    Serial.print(" ");
    Serial.print(steer[3], HEX);
    Serial.print(" ");
    Serial.print(steer[4], HEX);
    Serial.print(" ");
    Serial.print(steer[5], HEX);
    Serial.print(" ");
    Serial.print(steer[6], HEX);
    Serial.print(" ");
    Serial.println(steer[7], HEX);
    
    
    }

  }  
